from typing import Dict, List


class ConfigDefaults:
    OPTIMIZER_BUILDER: str
    MODEL_SERVING: bool
    EPOCHS: int
    SKIPEVAL: bool
    TEST_FREQUENCY: int
    TRAINER: str
    MODEL_VERSION: int
    MODEL_CORE_NAME: str
    MODEL_BACKBONE: str
    MODEL_QUALIFIER: str
    DRIVE_BACKUP: bool
    SAVE_FREQUENCY: int
    CHECKPOINT_DIRECTORY: str

    SHAPE: List[int]
    NORMALIZATION_MEAN: List[int]
    NORMALIZATION_STD: List[int]
    NORMALIZATION_SCALE: int
    H_FLIP: float
    T_CROP: bool
    RANDOM_ERASE: bool
    RANDOM_ERASE_VALUE: float
    CHANNELS: int
    BATCH_SIZE: int
    WORKERS: int

    BUILDER: str
    MODEL_ARCH: str
    MODEL_BASE: str
    MODEL_NORMALIZATION: str
    MODEL_KWARGS: Dict[str, str]

    LOSSES: List[str]
    LOSS_KWARGS: List[Dict[str, str]]
    LAMBDAS: List[int]
    LOSS_LABEL: str
    LOSS_NAME: str

    OPTIMIZER_NAME: str
    OPTIMIZER_KWARGS: Dict[str, str]
    BASE_LR: float
    LR_BIAS_FACTOR: float
    WEIGHT_DECAY: float
    WEIGHT_BIAS_FACTOR: float
    FP16: bool

    SCHEDULER_NAME: str
    LR_SCHEDULER: str
    LR_KWARGS: Dict[str, str]

    STEP_VERBOSE: int

    def __init__(self, **kwargs):

        self.OPTIMIZER_BUILDER = kwargs.get(
            "OPTIMIZER_BUILDER", "ClassificationOptimizer"
        )
        self.MODEL_SERVING = kwargs.get("MODEL_SERVING", False)
        self.EPOCHS = kwargs.get("EPOCHS", 10)
        self.SKIPEVAL = kwargs.get("SKIPEVAL", False)
        self.TEST_FREQUENCY = kwargs.get("TEST_FREQUENCY", 5)
        self.TRAINER = kwargs.get("TRAINER", "ClassificationTrainer")

        self.MODEL_VERSION = kwargs.get("MODEL_VERSION", "v1")
        self.MODEL_CORE_NAME = kwargs.get("MODEL_CORE_NAME", "model")
        self.MODEL_BACKBONE = kwargs.get("MODEL_BACKBONE", "res18")
        self.MODEL_QUALIFIER = kwargs.get("MODEL_QUALIFIER", "all")
        self.DRIVE_BACKUP = kwargs.get("DRIVE_BACKUP", False)
        self.SAVE_FREQUENCY = kwargs.get("SAVE_FREQUENCY", 5)
        self.CHECKPOINT_DIRECTORY = kwargs.get("CHECKPOINT_DIRECTORY", "checkpoint")

        self.SHAPE = kwargs.get("SHAPE", [100, 100])
        self.NORMALIZATION_MEAN = kwargs.get("NORMALIZATION_MEAN", 0.5)
        self.NORMALIZATION_STD = kwargs.get("NORMALIZATION_STD", 0.5)
        self.NORMALIZATION_SCALE = kwargs.get("NORMALIZATION_SCALE", 255)
        self.H_FLIP = kwargs.get("H_FLIP", 0)
        self.T_CROP = kwargs.get("T_CROP", False)
        self.RANDOM_ERASE = kwargs.get("RANDOM_ERASE", False)
        self.RANDOM_ERASE_VALUE = kwargs.get("RANDOM_ERASE_VALUE", 0.5)
        self.CHANNELS = kwargs.get("CHANNELS", 3)
        self.BATCH_SIZE = kwargs.get("BATCH_SIZE", 32)
        self.WORKERS = kwargs.get("WORKERS", 2)

        self.BUILDER = kwargs.get("BUILDER", "classification_model_builder")
        self.MODEL_ARCH = kwargs.get("MODEL_ARCH", "ClassificationResnet")
        self.MODEL_BASE = kwargs.get("MODEL_BASE", "resnet18")
        self.MODEL_NORMALIZATION = kwargs.get("MODEL_NORMALIZATION", "bn")
        self.MODEL_KWARGS = kwargs.get("MODEL_KWARGS", {})

        self.LOSSES = kwargs.get("LOSSES", [])
        self.LOSS_KWARGS = kwargs.get("LOSS_KWARGS", [])
        self.LAMBDAS = kwargs.get("LAMBDAS", [1.0])
        self.LOSS_LABEL = kwargs.get("LOSS_LABEL", "color")
        self.LOSS_NAME = kwargs.get("LOSS_NAME", "loss-1")

        self.OPTIMIZER_NAME = kwargs.get("OPTIMIZER_NAME", "opt-1")
        self.OPTIMIZER = kwargs.get("OPTIMIZER", "Adam")
        self.OPTIMIZER_KWARGS = kwargs.get("OPTIMIZER_KWARGS", {})
        self.BASE_LR = kwargs.get("BASE_LR", 0.001)
        self.LR_BIAS_FACTOR = kwargs.get("LR_BIAS_FACTOR", 1.0)
        self.WEIGHT_DECAY = kwargs.get("WEIGHT_DECAY", 0.0005)
        self.WEIGHT_BIAS_FACTOR = kwargs.get("WEIGHT_BIAS_FACTOR", 0.0005)
        self.FP16 = kwargs.get("FP16", False)

        self.SCHEDULER_NAME = kwargs.get("SCHEDULER_NAME", "opt-1")
        self.LR_SCHEDULER = kwargs.get("LR_SCHEDULER", "StepLR")
        self.LR_KWARGS = kwargs.get("LR_KWARGS", {"step_size": 20})

        self.STEP_VERBOSE = kwargs.get("STEP_VERBOSE", 100)



from setuptools import setup
import os

_PATH_ROOT = os.path.dirname(__file__)
_PATH_REQUIRE = os.path.join(_PATH_ROOT, "requirements")


requires = [
        "kaptan>=0.5.12",
        "scikit-learn>=1.0.2",
        "torch>=1.10.*",
        "torchinfo>=1.6.5",
        "torchvision>=0.11.*",
        "Pillow>=9.0.*",
        "tqdm>=4.63.*",
    ]


# Create full install that includes all extra dependencies

setup(
    name="ednaml",
    install_requires=requires,
    keywords=["deep learning", "pytorch", "AI"],
)


# Proposed splits for CUB200 from
# Zero-shot Learning - The Good, the Bad and the Ugly. Y. Xian, B. Schiele, Z. Akata. IEEE CVPR 2017.


trainval = r"""002.Laysan_Albatross
003.Sooty_Albatross
015.Lazuli_Bunting
016.Painted_Bunting
020.Yellow_breasted_Chat
022.Chuck_will_Widow
047.American_Goldfinch
048.European_Goldfinch
067.Anna_Hummingbird
068.Ruby_throated_Hummingbird
069.Rufous_Hummingbird
073.Blue_Jay
074.Florida_Jay
075.Green_Jay
076.Dark_eyed_Junco
089.Hooded_Merganser
090.Red_breasted_Merganser
100.Brown_Pelican
149.Brown_Thrasher
150.Sage_Thrasher
001.Black_footed_Albatross
014.Indigo_Bunting
034.Gray_crowned_Rosy_Finch
035.Purple_Finch
101.White_Pelican
120.Fox_Sparrow
110.Geococcyx
085.Horned_Lark
008.Rhinoceros_Auklet
054.Blue_Grosbeak
171.Myrtle_Warbler
041.Scissor_tailed_Flycatcher
056.Pine_Grosbeak
109.American_Redstart
187.American_Three_toed_Woodpecker
175.Pine_Warbler
030.Fish_Crow
195.Carolina_Wren
051.Horned_Grebe
107.Common_Raven
117.Clay_colored_Sparrow
135.Bank_Swallow
134.Cape_Glossy_Starling
046.Gadwall
147.Least_Tern
037.Acadian_Flycatcher
160.Black_throated_Blue_Warbler
126.Nelson_Sharp_tailed_Sparrow
137.Cliff_Swallow
125.Lincoln_Sparrow
027.Shiny_Cowbird
189.Red_bellied_Woodpecker
197.Marsh_Wren
186.Cedar_Waxwing
158.Bay_breasted_Warbler
064.Ring_billed_Gull
044.Frigatebird
007.Parakeet_Auklet
183.Northern_Waterthrush
142.Black_Tern
086.Pacific_Loon
159.Black_and_white_Warbler
081.Pied_Kingfisher
128.Seaside_Sparrow
011.Rusty_Blackbird
145.Elegant_Tern
029.American_Crow
166.Golden_winged_Warbler
059.California_Gull
095.Baltimore_Oriole
155.Warbling_Vireo
010.Red_winged_Blackbird
141.Artic_Tern
102.Western_Wood_Pewee
079.Belted_Kingfisher
106.Horned_Puffin
192.Downy_Woodpecker
005.Crested_Auklet
083.White_breasted_Kingfisher
178.Swainson_Warbler
162.Canada_Warbler
194.Cactus_Wren
196.House_Wren
050.Eared_Grebe
039.Least_Flycatcher
105.Whip_poor_Will
036.Northern_Flicker
032.Mangrove_Cuckoo
146.Forsters_Tern
082.Ringed_Kingfisher
060.Glaucous_winged_Gull
144.Common_Tern
199.Winter_Wren
093.Clark_Nutcracker
198.Rock_Wren
066.Western_Gull
099.Ovenbird
053.Western_Grebe
151.Black_capped_Vireo
018.Spotted_Catbird
152.Blue_headed_Vireo
116.Chipping_Sparrow
061.Heermann_Gull
025.Pelagic_Cormorant
024.Red_faced_Cormorant
078.Gray_Kingbird
017.Cardinal
176.Prairie_Warbler
058.Pigeon_Guillemot
021.Eastern_Towhee
193.Bewick_Wren
057.Rose_breasted_Grosbeak
040.Olive_sided_Flycatcher
153.Philadelphia_Vireo
088.Western_Meadowlark
013.Bobolink
118.House_Sparrow
121.Grasshopper_Sparrow
179.Tennessee_Warbler
062.Herring_Gull
154.Red_eyed_Vireo
092.Nighthawk
038.Great_Crested_Flycatcher
140.Summer_Tanager
182.Yellow_Warbler
096.Hooded_Oriole
172.Nashville_Warbler
071.Long_tailed_Jaeger
042.Vermilion_Flycatcher
185.Bohemian_Waxwing
177.Prothonotary_Warbler
019.Gray_Catbird
065.Slaty_backed_Gull
009.Brewer_Blackbird
112.Great_Grey_Shrike
063.Ivory_Gull
006.Least_Auklet
080.Green_Kingfisher
181.Worm_eating_Warbler
108.White_necked_Raven
122.Harris_Sparrow
115.Brewer_Sparrow
184.Louisiana_Waterthrush
167.Hooded_Warbler
129.Song_Sparrow
133.White_throated_Sparrow
114.Black_throated_Sparrow
170.Mourning_Warbler
131.Vesper_Sparrow
174.Palm_Warbler"""

query = r"""043.Yellow_bellied_Flycatcher
111.Loggerhead_Shrike
023.Brandt_Cormorant
098.Scott_Oriole
055.Evening_Grosbeak
130.Tree_Sparrow
139.Scarlet_Tanager
123.Henslow_Sparrow
156.White_eyed_Vireo
124.Le_Conte_Sparrow
200.Common_Yellowthroat
072.Pomarine_Jaeger
173.Orange_crowned_Warbler
028.Brown_Creeper
119.Field_Sparrow
165.Chestnut_sided_Warbler
103.Sayornis
180.Wilson_Warbler
077.Tropical_Kingbird
012.Yellow_headed_Blackbird
045.Northern_Fulmar
190.Red_cockaded_Woodpecker
191.Red_headed_Woodpecker
138.Tree_Swallow
157.Yellow_throated_Vireo
052.Pied_billed_Grebe
033.Yellow_billed_Cuckoo
164.Cerulean_Warbler
031.Black_billed_Cuckoo
143.Caspian_Tern
094.White_breasted_Nuthatch
070.Green_Violetear
097.Orchard_Oriole
091.Mockingbird
104.American_Pipit
127.Savannah_Sparrow
161.Blue_winged_Warbler
049.Boat_tailed_Grackle
169.Magnolia_Warbler
148.Green_tailed_Towhee
113.Baird_Sparrow
087.Mallard
163.Cape_May_Warbler
136.Barn_Swallow
188.Pileated_Woodpecker
084.Red_legged_Kittiwake
026.Bronzed_Cowbird
004.Groove_billed_Ani
132.White_crowned_Sparrow
168.Kentucky_Warbler"""


trainval = trainval.split("\n")
query = query.split("\n")



from torch import nn
from ednaml.models import ModelAbstract
from utils import layers
import torch

# CHANGELOG: secondary attention is List, not a single number


class ClassificationResnet(ModelAbstract):
    """Basic CoLabel Resnet model.

    A CoLabel model is a base ResNet, but during prediction, employs additional pieces such as 
    an ensemble voter, a heuristic based on the prediction output probabilities, as well as (if desired), 
    holistic nested side inputs.

    Args: (TODO)
        base (str): The architecture base for resnet, i.e. resnet50, resnet18
        weights (str, None): Path to weights file for the architecture base ONLY. If not provided, base initialized with random values.
        normalization (str, None): Can be None, where it is torch's normalization. Else create a normalization layer. Supports: ["bn", "l2", "in", "gn", "ln"]
        embedding_dimensions (int): Dimensions for the feature embedding. Leave empty if feature dimensions should be same as architecture core output (e.g. resnet50 base model has 2048-dim feature outputs). If providing a value, it should be less than the architecture core's base feature dimensions.
        softmax_dimensions (int, None): Number of FC dimensions for classification

    Kwargs (MODEL_KWARGS):
        last_stride (int, 1): The final stride parameter for the architecture core. Should be one of 1 or 2.
        attention (str, None): The attention module to use. Only supports ['cbam', None]
        input_attention (bool, false): Whether to include the IA module
        secondary_attention (List[int], None): Whether to modify CBAM to apply it to specific Resnet basic blocks. None means CBAM is applied to all. Otherwise, CBAM is applied only to the basic blocks provided here in List.
        part_attention (bool): Whether to use local attention

    Default Kwargs (DO NOT CHANGE OR ADD TO MODEL_KWARGS):
        zero_init_residual (bool, false): Whether the final layer uses zero initialization
        top_only (bool, true): Whether to keep only the architecture base without imagenet fully-connected layers (1000 classes)
        num_classes (int, 1000): Number of features in final imagenet FC layer
        groups (int, 1): Used during resnet variants construction
        width_per_group (int, 64): Used during resnet variants construction
        replace_stride_with_dilation (bool, None): Well, replace stride with dilation...
        norm_layer (nn.Module, None): The normalization layer within resnet. Internally defaults to nn.BatchNorm2D

    Methods: 
        forward: Process a batch

    """

    model_name = "ClassificationResNet"
    model_arch = "ClassificationResNet"
    number_outputs = 1
    output_classnames = ["out1"]
    softmax_dimensions = [2048]
    secondary_outputs = []

    def __init__(
        self, base="resnet50", weights=None, normalization=None, metadata=None, **kwargs
    ):
        super().__init__(
            base=base,
            weights=weights,
            normalization=normalization,
            metadata=metadata,
            **kwargs
        )

    def model_attributes_setup(self, **kwargs):
        self.embedding_dimensions = kwargs.get("embedding_dimensions", None)
        if self.normalization == "":
            self.normalization = None

        self.softmax_dimensions = [
            kwargs.get("softmax_dimensions", self.metadata.getLabelDimensions())
        ]

        self.output_classnames = [kwargs.get("output_classnames", "out1")]

        self.base = None
        self.gap = None
        self.emb_linear = None
        self.feat_norm = None
        self.softmax = None

    def model_setup(self, **kwargs):
        self.build_base(
            **kwargs
        )  # All kwargs are passed into build_base,, which in turn passes kwargs into _resnet()
        self.build_normalization()
        self.build_softmax()

    def build_base(self, **kwargs):
        """Build the model base.

        Builds the architecture base/core.
        """
        _resnet = __import__("backbones.resnet", fromlist=["resnet"])
        _resnet = getattr(_resnet, self.model_base)
        # Set up the resnet backbone
        self.base = _resnet(last_stride=1, **kwargs)
        if self.weights is not None:
            self.base.load_param(self.weights)

        self.gap = nn.AdaptiveAvgPool2d(1)

        self.emb_linear = torch.nn.Identity()
        if self.embedding_dimensions is None:
            self.embedding_dimensions = 512 * self.base.block.expansion
        if self.embedding_dimensions != 512 * self.base.block.expansion:
            self.emb_linear = nn.Linear(
                self.base.block.expansion * 512, self.embedding_dimensions, bias=False
            )

    def build_normalization(self):
        if self.normalization == "bn":
            self.feat_norm = nn.BatchNorm1d(self.embedding_dimensions)
            self.feat_norm.bias.requires_grad_(False)
            self.feat_norm.apply(self.weights_init_kaiming)
        elif self.normalization == "in":
            self.feat_norm = layers.FixedInstanceNorm1d(
                self.embedding_dimensions, affine=True
            )
            self.feat_norm.bias.requires_grad_(False)
            self.feat_norm.apply(self.weights_init_kaiming)
        elif self.normalization == "gn":
            self.feat_norm = nn.GroupNorm(
                self.embedding_dimensions // 16, self.embedding_dimensions, affine=True
            )
            self.feat_norm.bias.requires_grad_(False)
            self.feat_norm.apply(self.weights_init_kaiming)
        elif self.normalization == "ln":
            self.feat_norm = nn.LayerNorm(
                self.embedding_dimensions, elementwise_affine=True
            )
            self.feat_norm.bias.requires_grad_(False)
            self.feat_norm.apply(self.weights_init_kaiming)
        elif self.normalization == "l2":
            self.feat_norm = layers.L2Norm(self.embedding_dimensions, scale=1.0)
        elif self.normalization is None or self.normalization == "":
            self.feat_norm = torch.nn.Identity()
        else:
            raise NotImplementedError()

    def build_softmax(self, **kwargs):
        if self.softmax_dimensions[0] is not None:
            self.softmax = nn.Linear(
                self.embedding_dimensions, self.softmax_dimensions[0], bias=False
            )
            self.softmax.apply(self.weights_init_softmax)
        else:
            self.softmax = None  # TODO replace this with a zero compute layer that yields zero and has no_grad...

    def base_forward(self, x):
        features = self.gap(self.base(x))
        features = features.view(features.shape[0], -1)
        features = self.emb_linear(features)
        return features

    def forward_impl(self, x, **kwargs):
        features = self.base_forward(x)

        # if self.feat_norm is not None: <-- no need, identity
        features = self.feat_norm(features)

        soft_logits = None
        if self.softmax:
            soft_logits = self.softmax(features)
        return (
            soft_logits,
            features,
            [],
        )  # soft logits are the softmax logits we will use to for training. We can use features to store the historical probability????



import os.path as osp
import torch
from PIL import Image
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import DataLoader as TorchDataLoader

import torchvision.transforms as T


class CoLabelDataset(TorchDataset):
    def __init__(self, dataset, transform=None):
        self.dataset = dataset
        self.transform = transform

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return (
            self.transform(self.load(self.dataset[idx][0])),
            self.dataset[idx][1],
        )  # NOTE make this return a tuple for deploy (tuple of compressed), label

    def load(self, img):
        if not osp.exists(img):
            raise IOError("{img} does not exist in path".format(img=img))
        img_load = Image.open(img).convert("RGB")
        return img_load


class CoLabelIntegratedDatasetGenerator:
    def __init__(
        self,
        gpus,
        i_shape=(208, 208),
        normalization_mean=0.5,
        normalization_std=0.5,
        normalization_scale=1.0 / 255.0,
        h_flip=0.5,
        t_crop=True,
        rea=True,
        **kwargs
    ):
        """ Data generator for training and testing. Works with the VeriDataCrawler. Should work with any crawler working on VeRi-like data. Not yet tested with VehicleID. Only  use with VeRi.

        Generates batches of batch size CONFIG.TRANSFORMATION.BATCH_SIZE, with CONFIG.TRANSFORMATION.INSTANCE unique ids. So if BATCH_SIZE=36 and INSTANCE=6, then generate batch of 36 images, with 6 identities, 6 image per identity. See arguments of setup function for INSTANCE.

        Args:
            gpus (int): Number of GPUs
            i_shape (int, int): 2D Image shape
            normalization_mean (float): Value to pass as mean normalization parameter to pytorch Normalization
            normalization_std (float): Value to pass as std normalization parameter to pytorch Normalization
            normalization_scale (float): Value to pass as scale normalization parameter. Not used.
            h_flip (float): Probability of horizontal flip for image
            t_crop (bool): Whether to include random cropping
            rea (bool): Whether to include random erasing augmentation (at 0.5 prob)
        
        """

        self.gpus = gpus
        transformer_primitive = []
        transformer_primitive.append(T.Resize(size=i_shape))
        if h_flip > 0:
            transformer_primitive.append(T.RandomHorizontalFlip(p=h_flip))
        if t_crop:
            transformer_primitive.append(T.RandomCrop(size=i_shape))
        transformer_primitive.append(T.ToTensor())
        transformer_primitive.append(
            T.Normalize(mean=normalization_mean, std=normalization_std)
        )
        if rea:
            transformer_primitive.append(
                T.RandomErasing(
                    p=0.5, scale=(0.02, 0.4), value=kwargs.get("rea_value", 0)
                )
            )
        self.transformer = T.Compose(transformer_primitive)

        transformer_primitive_predict = []
        transformer_primitive_predict.append(T.Resize(size=i_shape))
        transformer_primitive_predict.append(T.ToTensor())
        transformer_primitive_predict.append(
            T.Normalize(mean=normalization_mean, std=normalization_std)
        )
        self.transformer_predict = T.Compose(transformer_primitive_predict)

    def setup(
        self,
        datacrawler,
        mode="train",
        batch_size=32,
        instance=8,
        workers=8,
        crawlerclass="color",
    ):
        """ Setup the data generator.

        Args:
            workers (int): Number of workers to use during data retrieval/loading
            datacrawler (VeRiDataCrawler): A DataCrawler object that has crawled the data directory
            mode (str): One of 'train', 'test', 'validation???'. 
        """

        if datacrawler is None:
            raise ValueError("Must pass DataCrawler instance. Passed `None`")
        self.workers = workers * self.gpus

        if mode == "train":
            self.__dataset = CoLabelDataset(
                datacrawler.metadata[mode]["crawl"], self.transformer
            )
        elif mode == "test":
            # For testing, we combine images in the query and testing set to generate batches
            self.__dataset = CoLabelDataset(
                datacrawler.metadata["val"]["crawl"]
                + datacrawler.metadata[mode]["crawl"],
                self.transformer_predict,
            )
        elif mode == "full":
            self.__dataset = CoLabelDataset(
                datacrawler.metadata["val"]["crawl"]
                + datacrawler.metadata["train"]["crawl"]
                + datacrawler.metadata["test"]["crawl"],
                self.transformer_predict,
            )
        else:
            raise NotImplementedError()

        if mode == "train":
            self.dataloader = TorchDataLoader(
                self.__dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
            self.num_entities = datacrawler.metadata[mode]["classes"][crawlerclass]
        elif mode == "test":
            self.dataloader = TorchDataLoader(
                self.__dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
            self.num_entities = datacrawler.metadata[mode]["classes"][crawlerclass]
        elif mode == "full":
            self.dataloader = TorchDataLoader(
                self.__dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
            self.num_entities = datacrawler.metadata[mode]["classes"][crawlerclass]
        else:
            raise NotImplementedError()

    def collate_simple(self, batch):
        img, class_id, = zip(
            *batch
        )  # NOTE for deploy, this is a tuple, label, i.e. (img, img, img), class
        class_id = torch.tensor(class_id, dtype=torch.int64)
        return (
            torch.stack(img, dim=0),
            class_id,
        )  # NOTE so this has to be modified...for tuple members...



from ednaml.utils.blocks.InputAttention import InputAttention
from ednaml.utils.blocks.DenseAttention import DenseAttention
from ednaml.utils.blocks.ChannelAttention import ChannelAttention
from ednaml.utils.blocks.SpatialAttention import SpatialAttention
from ednaml.utils.blocks.ResnetBasicBlock import ResnetBasicBlock
from ednaml.utils.blocks.ResnetBottleneck import ResnetBottleneck
from ednaml.utils.blocks.ResnetInput import ResnetInput



from torch import nn


class DenseAttention(nn.Module):  # Like spatial, but for all channels
    """Dense attention module for global attention from GLAMOR; see arXiv paper
    at https://arxiv.org/pdf/2002.02256.pdf
    """

    def __init__(self, planes):
        super(DenseAttention, self).__init__()
        self.dense_conv1 = nn.Conv2d(
            planes, planes, kernel_size=3, padding=1, bias=False
        )
        self.dense_relu1 = nn.LeakyReLU()
        self.dense_conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, padding=1, bias=False
        )
        self.dense_sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.dense_conv1(x)
        x = self.dense_relu1(x)
        x = self.dense_conv2(x)
        x = self.dense_sigmoid(x)
        return x



from typing import Dict

from ednaml.config.ConfigDefaults import ConfigDefaults


class OptimizerConfig:
    OPTIMIZER_NAME: str
    OPTIMIZER: str
    OPTIMIZER_KWARGS: Dict[str, str]
    BASE_LR: float
    LR_BIAS_FACTOR: float
    WEIGHT_DECAY: float
    WEIGHT_BIAS_FACTOR: float
    FP16: bool

    def __init__(self, optimizer_dict, defaults: ConfigDefaults):
        self.OPTIMIZER_NAME = optimizer_dict.get(
            "OPTIMIZER_NAME", defaults.OPTIMIZER_NAME
        )
        self.OPTIMIZER = optimizer_dict.get("OPTIMIZER", defaults.OPTIMIZER)
        self.OPTIMIZER_KWARGS = optimizer_dict.get(
            "OPTIMIZER_KWARGS", defaults.OPTIMIZER_KWARGS
        )
        self.BASE_LR = optimizer_dict.get("BASE_LR", defaults.BASE_LR)
        self.LR_BIAS_FACTOR = optimizer_dict.get(
            "LR_BIAS_FACTOR", defaults.LR_BIAS_FACTOR
        )
        self.WEIGHT_DECAY = optimizer_dict.get("WEIGHT_DECAY", defaults.WEIGHT_DECAY)
        self.WEIGHT_BIAS_FACTOR = optimizer_dict.get(
            "WEIGHT_BIAS_FACTOR", defaults.WEIGHT_BIAS_FACTOR
        )
        self.FP16 = optimizer_dict.get("FP16", defaults.FP16)
